@@ -32,9 +32,10 @@ from picard.track import Track
 from picard.script import ScriptParser
 from picard.ui.item import Item
 from picard.util import format_time, mbid_validate
+from picard.util.imagelist import ImageList, update_metadata_images
 from picard.util.textencoding import asciipunct
 from picard.cluster import Cluster
-from picard.collection import Collection, user_collections
+from picard.collection import add_release_to_user_collections
 from picard.mbxml import (
     release_group_to_metadata,
     release_to_metadata,
@@ -46,6 +47,11 @@ from picard.const import VARIOUS_ARTISTS_ID
 register_album_metadata_processor(coverart)
 
 
+class AlbumArtist(DataObject):
+    def __init__(self, id):
+        DataObject.__init__(self, id)
+
+
 class Album(DataObject, Item):
 
     release_group_loaded = QtCore.pyqtSignal()
@@ -53,6 +59,7 @@ class Album(DataObject, Item):
     def __init__(self, id, discid=None):
         DataObject.__init__(self, id)
         self.metadata = Metadata()
+        self.orig_metadata = Metadata()
         self.tracks = []
         self.loaded = False
         self.load_task = None
@@ -65,6 +72,8 @@ class Album(DataObject, Item):
         self.unmatched_files = Cluster(_("Unmatched Files"), special=True, related_album=self, hide_if_empty=True)
         self.errors = []
         self.status = None
+        self._album_artists = []
+        self.update_metadata_images_enabled = True
 
     def __repr__(self):
         return '<Album %s %r>' % (self.id, self.metadata[u"album"])
@@ -77,6 +86,20 @@ class Album(DataObject, Item):
             for file in self.unmatched_files.iterfiles():
                 yield file
 
+    def enable_update_metadata_images(self, enabled):
+        self.update_metadata_images_enabled = enabled
+
+    def append_album_artist(self, id):
+        """Append artist id to the list of album artists
+        and return an AlbumArtist instance"""
+        album_artist = AlbumArtist(id)
+        self._album_artists.append(album_artist)
+        return album_artist
+
+    def get_album_artists(self):
+        """Returns the list of album artists (as AlbumArtist objects)"""
+        return self._album_artists
+
     def _parse_release(self, document):
         log.debug("Loading release %r ...", self.id)
         self._tracks_loaded = False
@@ -123,13 +146,7 @@ class Album(DataObject, Item):
         m['totaldiscs'] = release_node.medium_list[0].count
 
         # Add album to collections
-        if "collection_list" in release_node.children:
-            for node in release_node.collection_list[0].collection:
-                if node.editor[0].text.lower() == config.setting["username"].lower():
-                    if node.id not in user_collections:
-                        user_collections[node.id] = \
-                            Collection(node.id, node.name[0].text, node.release_list[0].count)
-                    user_collections[node.id].releases.add(self.id)
+        add_release_to_user_collections(release_node)
 
         # Run album metadata plugins
         try:
@@ -209,26 +226,31 @@ class Album(DataObject, Item):
                 mm = Metadata()
                 mm.copy(self._new_metadata)
                 medium_to_metadata(medium_node, mm)
-                totalalbumtracks += int(mm["totaltracks"])
+                discpregap = False
 
                 for dj in djmix_ars.get(mm["discnumber"], []):
                     mm.add("djmixer", dj)
 
                 if "pregap" in medium_node.children:
-                    track = self._finalize_loading_track(medium_node.pregap[0], mm, artists, va, absolutetracknumber)
+                    discpregap = True
+                    absolutetracknumber += 1
+                    track = self._finalize_loading_track(medium_node.pregap[0], mm, artists, va, absolutetracknumber, discpregap)
                     track.metadata['~pregap'] = "1"
 
-                for track_node in medium_node.track_list[0].track:
-                    absolutetracknumber += 1
-                    track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber)
+                tracklist_node = medium_node.track_list[0]
+                track_count = int(tracklist_node.count)
+                if track_count:
+                    for track_node in tracklist_node.track:
+                        absolutetracknumber += 1
+                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
 
                 if "data_track_list" in medium_node.children:
                     for track_node in medium_node.data_track_list[0].track:
                         absolutetracknumber += 1
-                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber)
+                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
                         track.metadata['~datatrack'] = "1"
 
-            totalalbumtracks = str(totalalbumtracks)
+            totalalbumtracks = str(absolutetracknumber)
 
             for track in self._new_tracks:
                 track.metadata["~totalalbumtracks"] = totalalbumtracks
@@ -238,27 +260,29 @@ class Album(DataObject, Item):
             self._tracks_loaded = True
 
         if not self._requests:
+            self.enable_update_metadata_images(False)
             # Prepare parser for user's script
-            if config.setting["enable_tagger_script"]:
-                script = config.setting["tagger_script"]
-                if script:
-                    parser = ScriptParser()
-                    for track in self._new_tracks:
-                        # Run tagger script for each track
+            if config.setting["enable_tagger_scripts"]:
+                for s_pos, s_name, s_enabled, s_text in config.setting["list_of_scripts"]:
+                    if s_enabled and s_text:
+                        parser = ScriptParser()
+                        for track in self._new_tracks:
+                            # Run tagger script for each track
+                            try:
+                                parser.eval(s_text, track.metadata)
+                            except:
+                                self.error_append(traceback.format_exc())
+                            # Strip leading/trailing whitespace
+                            track.metadata.strip_whitespace()
+                        # Run tagger script for the album itself
                         try:
-                            parser.eval(script, track.metadata)
+                            parser.eval(s_text, self._new_metadata)
                         except:
                             self.error_append(traceback.format_exc())
-                        # Strip leading/trailing whitespace
-                        track.metadata.strip_whitespace()
-                    # Run tagger script for the album itself
-                    try:
-                        parser.eval(script, self._new_metadata)
-                    except:
-                        self.error_append(traceback.format_exc())
-                    self._new_metadata.strip_whitespace()
+                        self._new_metadata.strip_whitespace()
 
             for track in self.tracks:
+                track.metadata_images_changed.connect(self.update_metadata_images)
                 for file in list(track.linked_files):
                     file.move(self.unmatched_files)
             self.metadata = self._new_metadata
@@ -268,6 +292,7 @@ class Album(DataObject, Item):
             self.loaded = True
             self.status = None
             self.match_files(self.unmatched_files.files)
+            self.enable_update_metadata_images(True)
             self.update()
             self.tagger.window.set_statusbar_message(
                 N_('Album %(id)s loaded: %(artist)s - %(album)s'),
@@ -282,7 +307,7 @@ class Album(DataObject, Item):
                 func()
             self._after_load_callbacks = []
 
-    def _finalize_loading_track(self, track_node, metadata, artists, va, absolutetracknumber):
+    def _finalize_loading_track(self, track_node, metadata, artists, va, absolutetracknumber, discpregap):
         track = Track(track_node.recording[0].id, self)
         self._new_tracks.append(track)
 
@@ -297,6 +322,8 @@ class Album(DataObject, Item):
         artists.add(tm["artist"])
         if va:
             tm["compilation"] = "1"
+        if discpregap:
+            tm["~discpregap"] = "1"
 
         # Run track metadata plugins
         try:
@@ -364,10 +391,14 @@ class Album(DataObject, Item):
     def _add_file(self, track, file):
         self._files += 1
         self.update(update_tracks=False)
+        file.metadata_images_changed.connect(self.update_metadata_images)
+        self.update_metadata_images()
 
     def _remove_file(self, track, file):
         self._files -= 1
         self.update(update_tracks=False)
+        file.metadata_images_changed.disconnect(self.update_metadata_images)
+        self.update_metadata_images()
 
     def match_files(self, files, use_recordingid=True):
         """Match files to tracks on this album, based on metadata similarity or recordingid."""
@@ -437,7 +468,7 @@ class Album(DataObject, Item):
         return True
 
     def can_view_info(self):
-        return (self.loaded and self.metadata and self.metadata.images) or self.errors
+        return (self.loaded and (self.metadata.images or self.orig_metadata.images)) or self.errors
 
     def is_album_like(self):
         return True
@@ -459,8 +490,10 @@ class Album(DataObject, Item):
         if not self.tracks:
             return False
         for track in self.tracks:
-            if track.num_linked_files != 1:
+            if not track.is_complete():
                 return False
+        if self.get_num_unmatched_files():
+            return False
         else:
             return True
 
@@ -499,8 +532,21 @@ class Album(DataObject, Item):
                 unsaved = self.get_num_unsaved_files()
                 if unsaved:
                     text += '; %d*' % (unsaved,)
-                text += ungettext("; %i image", "; %i images",
-                                  len(self.metadata.images)) % len(self.metadata.images)
+                # CoverArt.set_metadata uses the orig_metadata.images if metadata.images is empty
+                # in order to show existing cover art if there's no cover art for a release. So
+                # we do the same here in order to show the number of images consistently.
+                if self.metadata.images:
+                    metadata = self.metadata
+                else:
+                    metadata = self.orig_metadata
+
+                number_of_images = len(metadata.images)
+                if getattr(metadata, 'has_common_images', True):
+                    text += ungettext("; %i image", "; %i images",
+                                      number_of_images) % number_of_images
+                else:
+                    text += ungettext("; %i image not in all tracks", "; %i different images among tracks",
+                                      number_of_images) % number_of_images
                 return text + ')'
             else:
                 return title
@@ -532,6 +578,23 @@ class Album(DataObject, Item):
             self.tagger.albums[mbid] = self
             self.load(priority=True, refresh=True)
 
+    def update_metadata_images(self):
+        if not self.update_metadata_images_enabled:
+            return
+
+        update_metadata_images(self)
+
+        self.update(False)
+
+    def keep_original_images(self):
+        self.enable_update_metadata_images(False)
+        for track in self.tracks:
+            track.keep_original_images()
+        for file in list(self.unmatched_files.files):
+            file.keep_original_images()
+        self.enable_update_metadata_images(True)
+        self.update_metadata_images()
+
 
 class NatAlbum(Album):
 
